10 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
20 ' renumber.bas
30 ' Geoff Graham 2010    http://geoffg.net
35 ' updated by James Deakins January 2013 to use newer functions 
40 ' of MMBasic V4.3
45 '
50 ' This program will renumber a basic program including references
60 ' to line numbers in statements.  It is written for MMBasic.
96 '
97 ' For comparison, Geoff's original program was 218 lines.  This is well 
98 ' over 500 lines.
99 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
100 '
110 false = 0 : true = NOT false
120 defaultStart = 1 : defaultFirst = 10 : defaultIncrement = 10
130 ' 
140 MaxParams = 20
150 DIM Param$(MaxParams)   ' the parameters after a keyword 
160 DIM ParamPtr(MaxParams) ' the pointer position of start of parameters 
170 '
180 PRINT "This program will renumber a BASIC program including references"
190 PRINT "to line numbers in GOSUB, GOTO, etc.  Minimal error checking"
200 PRINT "is done so the program must run correctly before renumbering."
210 PRINT
220 INPUT "Program to renumber: ", iname$
225 If iname$ = "" then iname$ = "test2.bas"
230 LinesInFile = CountLines(iname$) 
232 DIM lines(LinesInFile)
234 DIM labelLines(LinesInFile)       'TODO use it
235
236 OPEN iname$ FOR input  AS #1 
237 '
240 INPUT "   Output file name: ", oname$
245 If oname$ = "" then oname$ = "renum-ed.bas"
250 OPEN oname$ FOR output AS #2
260 '
' 270 PRINT : PRINT "In the input file, what is the desired:"
' 275 PRINT "(Default"; defaultStart; ")"; :     INPUT "  Line num where renumbering is to start (Enter key for default)"; In$
' 280 If In$ = "" then start = defaultStart else start = VAL(In$)
285 PRINT : PRINT "In the output file, what is the desired:"
290 PRINT "(Default"; defaultFirst; ")"; :     INPUT " First num to be used in the new sequence (Enter key for default)"; In$
295 If In$ = "" then first = defaultFirst else first = VAL(In$)
300 PRINT "(Default"; defaultIncrement; ")"; : INPUT "                 Increment for each line (Enter key for default)"; In$
302 If In$ = "" then increment = defaultIncrement else increment = VAL(In$)
304 '
305 'Get the length of the max line number so we can get neat spacing
306 MaxLineNum = first +((LinesInFile - 1)*increment)
307 MaxLineNum$ = Str$(MaxLineNum)
308 MaxLineNumLen = len(MaxLineNum$)
310 '
320 '
330 '
340 PRINT : PRINT "Checking for some of the potential problems.... "
350 'Pass one of the file - store the line numbers as integers in lines array
351 ' -------------------
352 ' Warn the user if any lines are out of order (an easy source of errors!)
353 ' Warn if any old style remarks signified by REM might be used.
354 PrevNum = 0
355 bLinesInOrder = True
356 bREMstatements = False
360 FOR LineNum = 0 to LinesInFile - 1 
370   Line INPUT #1, inp$
375   ThisNum = VAL(inp$)
380   lines(LineNum) = ThisNum
385   If (ThisNum <= PrevNum) AND (ThisNum) <> 0) Then ' a blank line number (allowable) will return zero
386     PRINT "File line "; LineNum + 1; " has its program linenumber ("; ThisNum; ") out of order!"
387     bLinesInOrder = False
388   Endif
390   UCLine$ = UCASE$(inp$)
395   if INSTR(1, UCLine$, "REM ") > 0 Then bREMstatements = True
400   PrevNum = ThisNum
405 Next LineNum
410 CLOSE #1
415 '
420 '
424 if NOT bLinesInOrder Then
425   ? "Duplicate line numbers found, or lines out of order. Only the first of a run"
426   ? "is identified, so also check the line numbers below the ones identified."
428   ?
429   Input "Line errors found.  Continue (Y/N)"; In$
430   if Ucase$(In$) = "N" then END
431 endif
432 '
    if bREMstatements = True Then
      ?
      ? "WARNING: Old style REM statements may have been found (REM is "
      ? "the old way of signifying remarks).  This program identifies "
      ? "REM statements by searching for 'REM ', so be careful using "
      ? "variable names ending in rem, or words ending in rem, like "
      ? "Harem, Spirem, Theorem or Millirem in your program.  They will "
      ? "be mistaken for Remarks  and stop the renumbering working "
      ? "correctly on that line."
      ?
      INPUT "Do you want to continue (Y/N)"; In$
      if Ucase$(In$) = "N" Then END
    endif

433 'Pass 2 of the file - renumber to output file
434 ' -----------------
435 PRINT : PRINT "Writing to Output File: "
440 OPEN iname$ FOR INPUT AS #1  
450 FOR lineindex = 0 TO LinesInFile - 1  
460   LINE INPUT #1, inp$ 
470   ProcessedLine$ = ProcessLine$(inp$, lineindex)
480   PRINT #2, LEFT$(ProcessedLine$, LEN(ProcessedLine$)) 
490   PRINT ".";  
500 NEXT lineindex   
510 '  
520 CLOSE #1, #2
530 PRINT  
540 PRINT "Finished, processed";LinesInFile;" lines"  
550 END  
560 '  
570 '
580 '
590 '  
600 ''''''''''''''''''''''''''''''''''''''''''''  
610 'subroutines and functions 
620 ''''''''''''''''''''''''''''''''''''''''''''  
700 '  
710 '  
720 '  
730 ''''''''''''''''''''''''''''''''''''''''''''  
740 'Function  
750 'process a line (in pInp$) and replace all  
760 'line numbers.  returns with pInp$ updated  
770 ''''''''''''''''''''''''''''''''''''''''''''
Function ProcessLine$(pInp$, pLineIndex)  
Local original$, newptr
'
original$ = pInp$ 
ptr = 1

' move pointer past any leading spaces
ptr = MovePtrPastSpaces(pInp$, ptr)

' move pointer past line number
ptr = MovePtrPastLineNumber(pInp$, ptr, LineNum) ' linenumber is returned in LineNum

'insert the new line number.  
If lines(pLineIndex) > 0 THEN ' it has a line number
  NewLineNum = first +(lineindex*increment)
  NewLineNumLength = len(str$(NewLineNum)) 
  Indent$ = space$(MaxLineNumLen - NewLineNumLength)
  newline$ = STR$(NewLineNum)+ Indent$ + MID$(inp$, ptr)
Else ' no line number
  n = INSTR(ptr, inp$, ":")
  if n > 0 then ' could be a label without a line number
    bLabel = True
    for i = ptr to n 
      ch$ = MID$(inp$, i, 1)
      if (ch$ = " ") then bLabel = False ' a label is a string immediately followed by a colon
    next i
    if bLabel then
      newLine$ = mid$(inp$, ptr) ' output with no indent (ie, at start of line)
    else     
      newline$ = space$(MaxLineNumLen + ptr) + MID$(inp$, ptr) ' ? "colon in multi statement.  Not a label. 
    endif
  else   
    newline$ = space$(MaxLineNumLen + ptr) + MID$(inp$, ptr)  ' ? "no colon, no line number"
  endif
Endif  

UCnewline$ = UCASE$(newline$)
' A line can be made up of code and remarks.  Don't search for keywords in remarks.
' Separate the remarks from the code, and just process the code from now on.
' We will reattach the remarks after the code is processed.

ptrSingleQuote = INSTR(1, UCnewline$, "'")
ptrREM         = INSTR(1, UCnewline$, "REM ") 

' look for the start of remarks. Could be a single quote (') or REM.  If both, which is first?
ptrRemStart = 0
if ptrSingleQuote > 0 Then
  ptrRemStart = ptrSingleQuote
endif
if ptrREM > ptrRemStart Then
  ptrRemStart = ptrREM
endif

If ptrRemStart > 0 Then
  Remarks$ = MID$(newline$, ptrRemStart)
  Code$ = Left$(newline$, ptrRemStart - 1)
else  'it is all code 
  Remarks$ = ""
  Code$ = newline$ 
endif

' A code line may be made up of multiple statements separated by colons (:)
' Process each statement separately, and reassemble.
startOfStatement = 1
ProcessedLine$ = ""
DO 
    nextColon = INSTR(startOfStatement, Code$, ":")
    if nextColon = 0 then 
      endOfStatement = len(Code$)
    else
      endOfStatement = nextColon
    endif
    ' extract next statement and process it
    Statement$ = Mid$(Code$, startOfStatement, endOfStatement - startOfStatement + 1)
    ProcessedLine$ = ProcessedLine$ + ProcessStatement$(Statement$)
    startOfStatement = nextColon + 1 
Loop Until nextColon = 0

ProcessedLine$ = ProcessedLine$ + Remarks$   ' reattach remarks

ProcessLine$ = ProcessedLine$

End Function ' ProcessLine$
'
'''''''''''''''''''''''''''''''''''''''''''''''
' Function
' Process the statement
' Return processed statement
'''''''''''''''''''''''''''''''''''''''''''''''
' Note: the original code by Geoff had separate
' processing for the ON command. The ON processing
' has now been rolled into the same code as GOTO
' and GOSUB.
' GOTO and GOSUB will handle both a single line
' number or label, or multiple ones separated by 
' commas.
'''''''''''''''''''''''''''''''''''''''''''''''
'
Function ProcessStatement$(pStatement$)

Local newStatement$, UCstatement$, kwPtr ' points to posn of keyword in statement

DATA "THEN","GOTO","GOSUB","ELSE","SETPIN","SETTICK","" 

UCstatement$ = UCase$(pStatement$)
' Note: UCstatement$ and pStatement$ are both the same length - only difference is one is Upper Case.

RESTORE ' reset data read pointer to start of list  
bProcessedKeyword = false
newStatement$ = pStatement$

Read kw$
DO WHILE (kw$ <> "")                         ' null means end of list   
  kwPtr = 1                                  ' start at first character of statement
  DO
    kwPtr = INSTR(kwPtr, UCstatement$, kw$)  'find the keyword  
    IF kwPtr <> 0 THEN  
       kwPtr = kwPtr + LEN(kw$)              'step over it and process
      newStatement$ = ProcessKeyWord$(kwPtr, newStatement$, kw$)
      bProcessedKeyword = True
    ENDIF  
  LOOP WHILE kwPtr <> 0
  Read kw$ 
LOOP

If bProcessedKeyword = False then ' no keywords found
  ' "No Keywords found.  Leave as is"
  newStatement$ = pStatement$ 
endif

ProcessStatement$ = newStatement$
End Function  ' ProcessStatement$      
'
'''''''''''''''''''''''''''''''''''''''''''''
' Function
' Function is passed a pointer to where the
' nominated keyword is in the nominated statement.
' Return processed statement and revised pointer.
'
'''''''''''''''''''''''''''''''''''''''''''''
' NOTE: old form of IF allowed linenumber directly 
' after THEN and ELSE, but had to all be on one line.
' Otherwise the GOTO linenumber and GOSUB linenumber 
' processing would handle the IF THEN ELSE statement.
'''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Function ProcessKeyWord$(pPtr, pStatement$, pKw$) 

Local ptr, updatedPtr

updatedPtr = pPtr ' return value

' move past space/s
ptr = MovePtrPastSpaces(pStatement$, pPtr)

params = parse(pStatement$, ptr)

if pKw$ = "THEN" OR pKw$ = "ELSE" then ' for obsolete form - Then or Else immediately followed by line num or label
 
 P1$ = param$(1)
  endP1Ptr = paramPtr(1) + LEN(param$(1))
  if params = 0 then ' it's continued on the next line, so it can't be the obsolete form
    updatedStatement$ = pStatement$ ' so leave it as it is and the GOTO or GOSUB processing will handle it
    updatedPtr = LEN(Statement$)
  elseif ( UCASE$(P1$) = "GOTO" ) OR ( UCASE$(P1$) = "GOSUB" ) then
    ' "Next term is GOTO or GOSUB, so leaving as is"
    updatedStatement$ = pStatement$ ' gosub and goto processing will handle it
    ' leave pointer and statement as is
  elseif VAL(P1$) = 0 Then ' a label
    ' Next term is a label, so leaving it as is
    updatedStatement$ = pStatement$
   updatedPtr = endP1Ptr
  else
    ' "Updating obsolete form of Then or Else"
    newNumber$ = LookupLine$(P1$)
    updatedStatement$ = Left$(pStatement$, paramPtr(1) - 1) + newNumber$ + MID$(pStatement$, paramPtr(1) + LEN(P1$))
    updatedPtr = endP1Ptr
  endif  
    
elseif pKw$ = "GOTO" OR pKw$ = "GOSUB" then 

  ProcessMultipleTargets(pStatement$, ptr, params)
  updatedStatement$ = pStatement$
  updatedPtr = ptr ' pass back up the call chain

elseif pKw$ = "SETPIN" then  ' param 3 optional
  
  if params = 2 then
    updatedStatement$ = pStatement$ ' doesn't use optional label or line number
    updatedPtr = paramPtr(2) + LEN(param$(2))
  else
    P3$ = param$(3) ' a label or line number
    endP3Ptr = paramPtr(3) + LEN(P3$)
    if VAL(P3$) = 0 then
      updatedStatement$ = pStatement$ ' a label
    else
      newNumber$ = LookupLine$(P3$)
      updatedStatement$ = Left$(pStatement$, paramPtr(3) - 1) + newNumber$ + MID$(pStatement$, endP3Ptr)
    endif
    updatedPtr = endP3Ptr
  endif 
    
elseif pKw$ = "SETTICK" then ' param 2 mandatory
' do not change if target (param 2) is 0, ie settick 0,0 but otherwise change the second parameter 
  
  if params = 2 then
    P2$ = param$(2)
    endP2Ptr = paramPtr(2) + LEN(P2$)
    if P2$ = "0" then
      updatedStatement$ = pStatement$ ' reset of SetTick interrupt.  Don't change!
    elseif VAL(P2$) = 0 then
      updatedStatement$ = pStatement$ ' a label
    else
      newNumber$ = LookupLine$(P2$)
      updatedStatement$ = Left$(pStatement$, paramPtr(2) - 1) + newNumber$ + MID$(pStatement$, endP2Ptr)
    endif
    updatedPtr = endP2Ptr
  else
      ? "Malformed SETTICK command: " + pStatement$
      INPUT "Press Ctrl-C to stop, any other key to continue "; In$
  endif
else 
  "You should never get this message.  The program tried to process the unknown keyword "; pKw$
  Input "Press Ctrl-C to break, or any other key to continue"; In$
endif

pPtr = updatedPtr
ProcessKeyWord$ = updatedStatement$ '   
 
End Function  ' ProcessKeyWord$
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Subroutine 
' This subroutine processes 1 or more labels or linenumbers separated by commas 
' and optional spaces following a GOTO or GOSUB keyword but it also processes 
' the normal single label / linenumber calls like if x = 3 then goto 10 and, through 2
' separate calls to this subroutine, calls like If x = 3 then goto 10 else goto 20.
'
' enters the routine with the pointer pointing at the first character of the first token
' and the Param$ array holds the token values.
' Returns two values, updated pStatement$ and ptr.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Sub ProcessMultipleTargets(pStatement$, ptr, params)

Local updatedStatement$, P$, n, newNumber$, bExit

updatedStatement$ = pStatement$

bExit = False
n = 0
DO While not bExit
  n = n + 1
  if n > params then
    bExit = True
  else
    P$ = param$(n) 
    PLen = LEN(P$) 
    endPPtr = paramPtr(n) + PLen
    if VAL(P$) = 0 Then  ' a label
      ' updatedStatement$ = pStatement$
    else
      newNumber$ = LookupLine$(P$)
      updatedStatement$ = Left$(updatedStatement$, paramPtr(n) - 1) + newNumber$ + MID$(updatedStatement$, endPPtr)   
    endif
    UCP$ = UCASE$(param$(n+1))
    if INSTR(1, "ELSETHENGOTOGOSUB", UCP$) > 0 then 
      bExit = True
    endif
  endif
Loop
ptr = endPPtr 
pStatement$ = updatedStatement$   
End Sub   ' ProcessMultipleTargets
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Function
' Parse the string into tokens based on spaces and commas
' as separators.  Store the tokens in the global Param$ array.
' Store the offset within the original line of the start of 
' each token into another global array, ParamPtr.
'
' Returns the number of parameters in the array
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Function Parse(pStatement$, pPtr) 
Local i, n, nextChar$

' Clear the array that will hold the parameters
For i = 0 to MaxParams
  Param$(i) = ""
Next i

StatementLen = Len(pStatement$)

ParamNum = 0  
n = pPtr
' skip any leading spaces
NextChar$ = mid$(pStatement$, n, 1)
Do While NextChar$ = " "
  n = n + 1
  NextChar$ = mid$(pStatement$, n, 1)
Loop

' At this point we have skipped all the spaces (if any) after the keyword and are pointing at the first parameter

Do While n <= StatementLen
  Parameter$ = ""
  ParamNum = ParamNum + 1
  ParamPtr(ParamNum) = n
  
  Do While (NextChar$ <> " ") AND (NextChar$ <> ",") AND (n <= StatementLen)
    Parameter$ = Parameter$ + NextChar$
	n = n + 1
	NextChar$ = mid$(pStatement$, n, 1)
  Loop
  
  Param$(ParamNum) = Parameter$

    Do While ((NextChar$ = " ") OR (NextChar$ = ",")) AND (n <= StatementLen)
    n = n + 1
    NextChar$ = mid$(pStatement$, n, 1)
  Loop
Loop

Parse = ParamNum
end Function   ' Parse
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Function
' Search the "lines" array to find the original line number.
' Return the value of the new line number.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Function LookupLine$(OldLine$)
Local bFound, i, OldLine

bFound = False
i = 0
OldLine = VAL(OldLine$)
Do While (i <= LinesInFile) AND NOT bFound
   IF lines(i) = OldLine THEN 
     bFound = True  
   Else
     i = i + 1
   Endif
Loop

if NOT bFound then
  PRINT : PRINT "No matching line number " + NextNumber$ + " found in original program."
  End  
Endif
 '
LookupLine$ = STR$(first + (i * increment))
  
End Function   ' LookupLine$
2190 '
2200 '''''''''''''''''''''''''''''''''''''''''''''''''''''''
2210 'Function
2220 'Return the number of lines in the input file.
2230 '''''''''''''''''''''''''''''''''''''''''''''''''''''''
2240 'Note: Closes and reopens file so input pointer reset to start.
2250 '''''''''''''''''''''''''''''''''''''''''''''''''''''''
2260 Function CountLines(FileName$)
2270 local Count
2280 '
2285 OPEN FileName$ for INPUT as #3
2290 Count = 0
2300 DO WHILE NOT EOF(#3)
2310     Line Input #3, inp$
2320     Count = Count + 1
2330 LOOP
2340 Close #3
2350 
2360 CountLines = Count
2370 End Function   ' CountLines
2390 '
2400 '''''''''''''''''''''''''''''''''''''''''''''''''''''
2410 'Function
2420 'scan the line and move the pointer to the first 
2430 'character past any leading spaces
2440 '
2450 '''''''''''''''''''''''''''''''''''''''''''''''''''''
2455 ' 
2460 Function MovePtrPastSpaces(pStr$, pPtr)
2470 local ptr
2480
2490 ptr = pPtr
2500 DO WHILE mid$(pStr$,ptr,1) = " "
2510   ptr = ptr + 1
2520 LOOP
2530 MovePtrPastSpaces = ptr
2540 End Function   ' MovePtrPastSpaces
2570 '
2580 '''''''''''''''''''''''''''''''''''''''''''''''''''''
2590 'Function
2600 'scan the line and move the pointer to the first 
2610 'character past any leading spaces
2620 'note: function also updates LineNum parameter
2630 '''''''''''''''''''''''''''''''''''''''''''''''''''''
2635 ' 
2640 Function MovePtrPastLineNumber(pStr$, pPtr, LineNum)
2650 local ptr, nbrStr$
2660
2670 ptr=pPtr
2680 ' then get the number (if there)  
2690 nbrstr$ = ""  
2695 ch$ = MID$(pStr$,ptr,1)
2700 DO WHILE ch$ >= "0" AND ch$ <= "9"  
2710   ptr = ptr + 1
2720   nbrstr$ = nbrstr$ + ch$
2725   ch$ = MID$(pStr$,ptr,1)
2730 LOOP 
2740 LineNum = Val(nbrstr$) 
2750 MovePtrPastLineNumber = ptr
2760 End Function   ' MovePtrPastLineNumber
